#pragma once

#include "nodes/changed_max_points.hpp"
#include "nodes/combined_grade_average.hpp"
#include "nodes/combined_grade_list.hpp"
#include "nodes/dynexite_exam.hpp"
#include "nodes/first_try.hpp"
#include "nodes/grade_list.hpp"
#include "nodes/node.hpp"
#include "nodes/passing_limit.hpp"
#include "nodes/student_data.hpp"
#include "nodes/user_grade_boundaries.hpp"
#include "nodes/user_grade_config.hpp"
#include "nodes/user_point_config.hpp"

#include "util/enums.hpp"

#include <tuple>

#include <QDebug>

template <typename... ChangedNode>
void updateNodesIfNecessary() noexcept {}

template <typename... ChangedNodes, typename N, typename... Nodes>
void updateNodesIfNecessary(N &n,
                            Nodes &...nodes) noexcept { /* TODO: noexcept? */
    if constexpr (AnyIsPredecessorOfAny<std::tuple<ChangedNodes...>,
                                        std::tuple<N>>::value) {
        n.recalculate();
    }
    updateNodesIfNecessary<ChangedNodes...>(nodes...);
}

struct DAG {

    UserPointConfig mUserPointConfig{};
    UserGradeConfig mUserGradeConfig{};
    DynexiteExam mExam{};
    StudentData mStudents{};

    ChangedMaxPoints<GradingType::Closed> mChangedMaxPointsClosed{
        mUserPointConfig, mExam};
    ChangedMaxPoints<GradingType::Open> mChangedMaxPointsOpen{mUserPointConfig,
                                                              mExam};
    FirstTry mFirstTry{mExam, mStudents};

    PassingLimit<GradingType::Closed> mClosedPassingLimit{
        mChangedMaxPointsClosed, mFirstTry, mUserPointConfig};
    PassingLimit<GradingType::Open> mOpenPassingLimit{mChangedMaxPointsOpen,
                                                      mUserPointConfig};

    UserGradeBoundaries<GradingType::Closed> mClosedUserBoundaries{
        mUserPointConfig, mChangedMaxPointsClosed, mClosedPassingLimit};
    UserGradeBoundaries<GradingType::Open> mOpenUserBoundaries{
        mUserPointConfig, mChangedMaxPointsOpen, mOpenPassingLimit};

    GradeList<GradingType::Closed> mClosedGrades{mUserPointConfig,
                                                 mClosedUserBoundaries, mExam};
    GradeList<GradingType::Open> mOpenGrades{mUserPointConfig,
                                             mOpenUserBoundaries, mExam};

    CombinedGradeList mGrades{mChangedMaxPointsClosed, mChangedMaxPointsOpen,
                              mClosedGrades, mOpenGrades, mUserGradeConfig};

    CombinedGradeAverage mGradeAverage{mGrades};

    template <typename... ChangedNodes>
    void updateMemberNodesIfNecessary() noexcept /* TODO: noexcept? */
    {
        updateNodesIfNecessary<ChangedNodes...>(
            mUserPointConfig, mUserGradeConfig, mExam, mStudents,
            mChangedMaxPointsClosed, mChangedMaxPointsOpen, mFirstTry,
            mClosedPassingLimit, mOpenPassingLimit, mClosedUserBoundaries,
            mOpenUserBoundaries, mClosedGrades, mOpenGrades, mGrades,
            mGradeAverage);
    }

    DAG(DynexiteExam exam, StudentData students,
        const UserGradeConfig &gradeConfig, const UserPointConfig &pointConfig)
        : mUserPointConfig(pointConfig), mUserGradeConfig(gradeConfig),
          mExam(std::move(exam)), mStudents(std::move(students)) {
        updateMemberNodesIfNecessary<DynexiteExam, StudentData, UserGradeConfig,
                                     UserPointConfig>();
    }

    // copy constructor: bind reference to this object, not to the copied-from
    // object
    DAG(const DAG &other)
        : mUserPointConfig(other.mUserPointConfig),
          mUserGradeConfig(other.mUserGradeConfig), mExam(other.mExam),
          mStudents(other.mStudents) {
        updateMemberNodesIfNecessary<DynexiteExam, StudentData, UserGradeConfig,
                                     UserPointConfig>();
    }
    // todo: move constructor etc

    // helper functions for easier access
    [[nodiscard]] constexpr const GradeBoundary &
    getPassingLimit(const GradingType grading) const noexcept {
        switch (grading) {
        default:
            assert(false);
            [[fallthrough]];
        case GradingType::Closed:
            return mClosedPassingLimit.mPassingLimit;
        case GradingType::Open:
            return mOpenPassingLimit.mPassingLimit;
        }
    }
    [[nodiscard]] constexpr const EnumStorage<GradeBoundary, Grade> &
    getUserGradeBoundaries(const GradingType grading) const noexcept {
        switch (grading) {
        default:
            assert(false);
            [[fallthrough]];
        case GradingType::Closed:
            return mClosedUserBoundaries.mBoundaries;
        case GradingType::Open:
            return mOpenUserBoundaries.mBoundaries;
        }
    }
    [[nodiscard]] constexpr const QList<Grade> &
    getGradeList(const GradingType grading) const noexcept {
        switch (grading) {
        default:
            assert(false);
            [[fallthrough]];
        case GradingType::Closed:
            return mClosedGrades.mGrades;
        case GradingType::Open:
            return mOpenGrades.mGrades;
        case GradingType::Combined:
            return mGrades.mGrades;
        }
    }
};
