/**
 * \file csv_parser.hpp
 * \brief Declares functions to read and write csv data.
 */

#pragma once

#include "data/csv_dump.hpp"

#include <filesystem>
#include <istream>
#include <memory>
#include <ostream>
#include <string>

/**
 * \brief Reads the file at the given path into a \ref CsvDump if possible.
 *
 * Tries to open the file at the given path and throws an
 * `std::ios_base::failure` if that is not possible. Afterwards it checks if the
 * number of delimiters per line is consistent and throws an
 * `std::runtime_error` if that is not the case. Finally it stores the data in a
 * \ref CsvDump.
 *
 * It tries the following delimiters in order until one seems plausible:
 * `;` `,` `\t`
 *
 * \param[in] p     Path to the file.
 *
 * \exception std::ios_base::failure Bad bit set while reading the file
 * \exception std::invalid_argument  The file could not be opened
 * \exception std::runtime_error     The number of delimiters per line is
 * inconsistent.
 *
 * \return Returns a pointer to the extracted CSV-data at the given path.
 */
[[nodiscard]] std::unique_ptr<CsvDump> readCsv(const std::filesystem::path &p);

/**
 * \brief Reads the string into a \ref CsvDump if possible.
 *
 * See \ref readCsv(const std::filesystem::path) for details.
 */
[[nodiscard]] std::unique_ptr<CsvDump> readCsv(const std::string &s);

/**
 * \brief Reads the generic input stream into a \ref CsvDump if possible.
 *
 * See \ref readCsv(const std::filesystem::path) for details.
 */
[[nodiscard]] std::unique_ptr<CsvDump> readCsv(std::istream &in);

/**
 * \brief Exports the given csv dump to the given stream.
 */
void writeCsv(std::ostream &out, const CsvDump &c, char delim = ';');

/**
 * \brief Exports the given csv dump to a string.
 */
[[nodiscard]] std::string writeCsv(const CsvDump &c, char delim = ';');

/**
 * \brief Creates a file at the given location and writes the csv dump to it.
 *
 * Overwrites anything previously stored in the file.
 */
void writeCsv(const std::filesystem::path &p, const CsvDump &c,
              char delim = ';');
