#include "grade_tab.hpp"

#include "data/data_manager.hpp"

#include <QGroupBox>
#include <QTabWidget>

inline constexpr std::size_t openGradesIdx = 0;
inline constexpr std::size_t closedGradesIdx = 1;
inline constexpr std::size_t combinedGradesIdx = 2;
inline constexpr std::size_t allGradesIdx = 3;

inline constexpr std::size_t openPointsIdx = 0;
inline constexpr std::size_t closedPointsIdx = 1;
inline constexpr std::size_t combinedPointsIdx = 2;

GradeTab::GradeTab(const ViewSettingsWindow &viewSettings, QWidget *parent)
    : QWidget(parent), mScrews(new MinMaxPointScrews(this)),
      mGradeHistoTabs(new QTabWidget(this)), mGradeHistos(),
      mPointHistoTabs(new QTabWidget(this)), mPointHistos(),
      mLayout(new QGridLayout(this)) {

    mLayout->addWidget(mScrews, 0, 0);
    mLayout->addWidget(mGradeHistoTabs, 0, 1);
    mLayout->addWidget(mPointHistoTabs, 1, 0, 1, 2);

    createGradeHistograms();
    createPointHistograms(viewSettings);

    // 30% - 70% split
    mLayout->setColumnStretch(0, 3);
    mLayout->setColumnStretch(1, 7);
    mLayout->setRowStretch(0, 1);
    mLayout->setRowStretch(1, 1);

    connect(&DataManager::getInstance(), &DataManager::rawDataChanged, this,
            &GradeTab::pointsChanged);
}

void GradeTab::handleResize() const {
    const auto plotAreaGradeHistos =
        mGradeHistos[static_cast<std::size_t>(mGradeHistoTabs->currentIndex())]
            ->chart()
            ->plotArea();

    for (const auto &gradeHisto : mGradeHistos) {
        gradeHisto->handleResize(plotAreaGradeHistos);
    }

    const auto plotAreaPointHistos =
        mPointHistos[static_cast<std::size_t>(mPointHistoTabs->currentIndex())]
            ->chart()
            ->plotArea();
    for (const auto &pointHisto : mPointHistos) {
        pointHisto->handleResize(plotAreaPointHistos);
    }
}

void GradeTab::pointsChanged() {
    const auto cur = DataManager::getCurrent();

    if (!cur)
        return;

    const bool hasClosed =
        cur->mExam.mPointsMax[GradingType::Closed] > Points{0.};
    const bool hasOpen = cur->mExam.mPointsMax[GradingType::Open] > Points{0.};

    mScrews->setEnabled(GradingType::Closed, hasClosed);
    mScrews->setEnabled(GradingType::Open, hasOpen);

    // Only show multiple tabs if there were open AND closed tasks
    if (hasClosed && hasOpen) {
        mGradeHistoTabs->setTabVisible(static_cast<int>(openGradesIdx), true);
        mGradeHistoTabs->setTabVisible(static_cast<int>(closedGradesIdx), true);
        mGradeHistoTabs->setTabVisible(static_cast<int>(combinedGradesIdx),
                                       true);
        mGradeHistoTabs->setTabVisible(static_cast<int>(allGradesIdx), true);

        mPointHistoTabs->setTabVisible(static_cast<int>(openPointsIdx), true);
        mPointHistoTabs->setTabVisible(static_cast<int>(closedPointsIdx), true);
        mPointHistoTabs->setTabVisible(static_cast<int>(combinedPointsIdx),
                                       true);

    } else {
        mGradeHistoTabs->setTabVisible(static_cast<int>(openGradesIdx), false);
        mGradeHistoTabs->setTabVisible(static_cast<int>(closedGradesIdx),
                                       false);
        mGradeHistoTabs->setTabVisible(static_cast<int>(combinedGradesIdx),
                                       true);
        mGradeHistoTabs->setTabVisible(static_cast<int>(allGradesIdx), false);

        mPointHistoTabs->setTabVisible(static_cast<int>(openPointsIdx), false);
        mPointHistoTabs->setTabVisible(static_cast<int>(closedPointsIdx),
                                       false);
        mPointHistoTabs->setTabVisible(static_cast<int>(combinedPointsIdx),
                                       true);
    }
}

void GradeTab::createGradeHistograms() {
    mGradeHistos[openGradesIdx] = new GradeHistogram({GradingType::Open}, this);
    mGradeHistos[closedGradesIdx] =
        new GradeHistogram({GradingType::Closed}, this);
    mGradeHistos[combinedGradesIdx] =
        new GradeHistogram({GradingType::Combined}, this);
    mGradeHistos[allGradesIdx] = new GradeHistogram(
        {GradingType::Open, GradingType::Closed, GradingType::Combined}, this);
    mGradeHistoTabs->addTab(mGradeHistos[openGradesIdx],
                            "Histogram of open grades");
    mGradeHistoTabs->addTab(mGradeHistos[closedGradesIdx],
                            "Histogram of closed grades");
    mGradeHistoTabs->addTab(mGradeHistos[combinedGradesIdx],
                            "Histogram of resulting grades");
    mGradeHistoTabs->addTab(mGradeHistos[allGradesIdx], "Show all histograms");
    mGradeHistoTabs->setCurrentIndex(static_cast<int>(combinedGradesIdx));
}

void GradeTab::createPointHistograms(const ViewSettingsWindow &viewSettings) {
    mPointHistos[combinedPointsIdx] =
        new PointHistogram(GradingType::Combined, viewSettings, this);
    mPointHistos[openPointsIdx] =
        new PointHistogram(GradingType::Open, viewSettings, this);
    mPointHistos[closedPointsIdx] =
        new PointHistogram(GradingType::Closed, viewSettings, this);
    mPointHistoTabs->addTab(mPointHistos[openPointsIdx],
                            "Histogram of open points");
    mPointHistoTabs->addTab(mPointHistos[closedPointsIdx],
                            "Histogram of closed points");
    mPointHistoTabs->addTab(mPointHistos[combinedPointsIdx],
                            "Histogram of total points");
    mPointHistoTabs->setCurrentIndex(combinedPointsIdx);
}
