/**
 * \file main_window.cpp
 * \brief Definition of the \ref MainWindow member functions.
 */

#include "main_window.hpp"

#include "export_rwthonline.hpp"

#include "data/data_manager.hpp"
#include "file_io/yml_parser.hpp"
#include "util/grade.hpp"
#include "util/grade_boundary.hpp"
#include "util/heuristics.hpp"

#include <QAction>
#include <QClipboard>
#include <QFileDialog>
#include <QGuiApplication>
#include <QMenuBar>

#include <algorithm>
#include <cmath>

// Surround the given string evenly with spaces, until it has a size of `size`
QString padToSize(QString str, const int size) {
    const auto spacesLeft = static_cast<int>(
        std::max(0., std::floor(static_cast<double>(size - str.size()) / 2.)));
    const auto spacesRight = static_cast<int>(
        std::max(0., std::ceil(static_cast<double>(size - str.size()) / 2.)));

    str.prepend(QString(spacesLeft, ' '));
    str.append(QString(spacesRight, ' '));

    return str;
}

void printBoundaries(QString &str, const GradingType type, const bool hasType,
                     const GradeBoundaryStepSize stepSize,
                     const int decimalPlaces) {

    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    if (hasType) {
        str += QString::fromStdString(toString(type)) +
               " part:\n"
               "Grade --- Min. points --- Max. points\n";

        auto lastGradeBoundary =
            cur->mChangedMaxPointsOpen.mChangedMaxPointBoundary;

        for (const auto g : grades) {

            const auto minBoundary =
                toPoints(cur->getUserGradeBoundaries(type)[g], stepSize).p;

            // For max points of 1.0, we use exam max points, not changed max
            // points.
            // For other max points, we go one grade boundary step below the
            // next higher grade.
            const auto maxBoundary =
                g == Grade::grade1_0 ? cur->mExam.mPointsMax[type].p
                                     : toPoints(lastGradeBoundary, stepSize).p -
                                           as<double>(stepSize);

            str +=
                padToSize(toString(g).c_str(), 7) + ":" +
                padToSize(QLocale().toString(minBoundary, 'f', decimalPlaces),
                          15) +
                "-" +
                padToSize(QLocale().toString(maxBoundary, 'f', decimalPlaces),
                          13) +
                "\n";
            lastGradeBoundary = cur->getUserGradeBoundaries(type)[g];
        }

        str += "\n";
    }
}

void copyGradingSchemeToClipboard() {
    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    auto *clipboard = QGuiApplication::clipboard();

    const bool hasClosed =
        cur->mExam.mPointsMax[GradingType::Closed] > Points{0.};
    const bool hasOpen = cur->mExam.mPointsMax[GradingType::Open] > Points{0.};

    QString gradingSchemeText = "";

    const auto stepSize = cur->mUserPointConfig.mGradeBoundaryStepSize;
    const auto decimalPlaces = nrDecimalPlaces(stepSize);

    printBoundaries(gradingSchemeText, GradingType::Open, hasOpen, stepSize,
                    decimalPlaces);
    printBoundaries(gradingSchemeText, GradingType::Closed, hasClosed, stepSize,
                    decimalPlaces);

    if (!hasOpen && !hasClosed) {
        gradingSchemeText +=
            "Selected exam has neither open nor closed points.";
    }

    clipboard->setText(gradingSchemeText);
}

void exportD2R(QWidget *parent) {
    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    const auto exportLocationString = QFileDialog::getSaveFileName(
        parent, "Save data all-in-one (d2r)", "",
        "All-in-one (D2R) file (*.d2r);;All files (*.*)");

    if (!exportLocationString.isEmpty()) {
        writeYml(exportLocationString.toStdWString());
    }
}

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent), mInputWindow(*this, parent), mBoundaryWindow(parent),
      mViewSettings(parent), mStatisticsWindow(parent) /*,
       mGradeTab(mViewSettings), mTaskTab(), mStudentsTab()*/
{

    setWindowTitle("Dynexite2RWTHonline");

    setCentralWidget(&mTabs);
    // handleCommandLineArgs();

    mGradeTab = new GradeTab{mViewSettings};
    mTaskTab = new TaskTab{};
    mStudentsTab = new StudentsTab{};
    mTabs.addTab(mGradeTab, "Grading");
    mTabs.addTab(mTaskTab, "Tasks");
    mTabs.addTab(mStudentsTab, "Students");

    // Create handle for all windows for operations that need to iterate over
    // all of them
    mSubWindowPtrs[0] = &mInputWindow;
    mSubWindowPtrs[1] = &mBoundaryWindow;
    mSubWindowPtrs[2] = &mViewSettings;
    mSubWindowPtrs[3] = &mStatisticsWindow;

    // Set appropriate flags for all small openable windows: not maximizable,
    // stay on top, closable
    for (const auto &window : mSubWindowPtrs) {
        window->setWindowFlags(
            Qt::WindowStaysOnTopHint | Qt::WindowCloseButtonHint |
            Qt::WindowMinimizeButtonHint | Qt::MSWindowsFixedSizeDialogHint);
        window->setWindowFlags(mBoundaryWindow.windowFlags() &
                               ~Qt::WindowMaximizeButtonHint);
    }

    // start input file selection
    mInputWindow.show();

    createMenuBar();
}

void MainWindow::resetToInputFiles(const DynexiteExam &exam,
                                   const StudentData &students,
                                   const YmlSettings &settings) {

    UserPointConfig pc;
    pc.mGradeBoundaryStepSize = settings.mUserPointConfigStepSize;
    if (DataManager::loadNewExam(exam, students, UserGradeConfig{}, pc)) {
        DataManager::makeNewCurrent();

        DataManager::applySettings(settings);

        // close input window and show main window
        mInputWindow.hide();
        showMaximized();

        updateMenuItems();
    }
}

void MainWindow::menuNewFile() { mInputWindow.show(); }

void MainWindow::menuOpenBoundaries() { mBoundaryWindow.show(); }

void MainWindow::menuOpenViewSettings() { mViewSettings.show(); }

void MainWindow::menuOpenStatistics() { mStatisticsWindow.show(); }

void MainWindow::closeEvent(QCloseEvent *event) {
    for (const auto &window : mSubWindowPtrs) {
        window->hide();
    }

    QMainWindow::closeEvent(event);
}

void MainWindow::resizeEvent(QResizeEvent *event) {
    QMainWindow::resizeEvent(event);

    mGradeTab->handleResize();
}

void MainWindow::updateMenuItems() const {
    const auto nrUnknownStudents = getNrUnknownStudents();

    mExportUnknownStudentsAct->setText("Manual: Unknown Students (" +
                                       QString::number(nrUnknownStudents) +
                                       ")");
    mExportUnknownStudentsAct->setEnabled(nrUnknownStudents > 0);
}

void MainWindow::createMenuBar() {

    /***************** Exam **************/
    mFileMenu = menuBar()->addMenu("&Exam");

    // "New": loads a new exam
    {
        const auto newAct = new QAction("&Open", this);
        newAct->setShortcuts(QKeySequence::Open);
        connect(newAct, &QAction::triggered, this, &MainWindow::menuNewFile);
        mFileMenu->addAction(newAct);
    }

    // "Options": change the required points for e.g. 2.7, change requirePassAll
    {
        const auto boundaryAct = new QAction("Options", this);
        boundaryAct->setShortcuts(QKeySequence::Preferences);
        connect(boundaryAct, &QAction::triggered, this,
                &MainWindow::menuOpenBoundaries);
        mFileMenu->addAction(boundaryAct);
    }

    // "Statistics": show more detailed statistics like number of attended
    // students
    {
        const auto statisticsAct = new QAction("Statistics", this);
        connect(statisticsAct, &QAction::triggered, this,
                &MainWindow::menuOpenStatistics);
        mFileMenu->addAction(statisticsAct);
    }

    // "Tasks": show more detailed task info (temporarily here)
    //{
    //    const auto tasksAct = new QAction("Tasks", this);
    //    connect(tasksAct, &QAction::triggered, this,
    //            &MainWindow::menuOpenTasks);
    //    mFileMenu->addAction(tasksAct);
    //}

    /***************** View **************/
    mViewMenu = menuBar()->addMenu("&View");

    // "View settings": change visual options
    {
        const auto viewSettingsAct = new QAction("Settings", this);
        connect(viewSettingsAct, &QAction::triggered, this,
                &MainWindow::menuOpenViewSettings);
        mViewMenu->addAction(viewSettingsAct);
    }

    /**************** Export *************/
    mExportMenu = menuBar()->addMenu("&Export");

    // Export grading scheme to clipboard
    {
        mExportGradingMenu = mExportMenu->addMenu("Grading scheme");
        const auto exportGradingSchemeClipboardAct =
            new QAction("To Clipboard", this);
        connect(exportGradingSchemeClipboardAct, &QAction::triggered,
                copyGradingSchemeToClipboard);
        mExportGradingMenu->addAction(exportGradingSchemeClipboardAct);
    }

    // Export students
    {
        mExportStudentsMenu = mExportMenu->addMenu("Results");
        // Export known students as CSV
        {
            const auto exportKnownStudentsAct =
                new QAction("RWTHonline: Known Students", this);
            connect(exportKnownStudentsAct, &QAction::triggered,
                    [this] { exportKnownStudents(this); });
            mExportStudentsMenu->addAction(exportKnownStudentsAct);
        }

        // Export unknown students as CSV
        {
            mExportUnknownStudentsAct =
                new QAction("Manual: Unknown Students", this);
            connect(mExportUnknownStudentsAct, &QAction::triggered,
                    [this] { exportUnknownStudents(this); });
            mExportStudentsMenu->addAction(mExportUnknownStudentsAct);
        }
    }

    // Export to d2r
    {
        const auto exportD2RAct = new QAction("D2R", this);
        connect(exportD2RAct, &QAction::triggered, [this] { exportD2R(this); });
        mExportMenu->addAction(exportD2RAct);
    }
}

/* void MainWindow::handleCommandLineArgs() const {
    // get list of command line arguments
    QStringList fileNames = QCoreApplication::arguments();

    // skip first argument (path of executable)
    if (fileNames.size() > 1) {
        mD2RFileChooserTab->newInputFile(fileNames[1]);
    }
}*/
