/**
 * \file point_histogram.cpp
 */

#include "point_histogram.hpp"

#include "view/main_window/view_settings.hpp"

#include "data/data_manager.hpp"

#include <QBarSet>
#include <QChart>
#include <QColor>
#include <QFontMetrics>
#include <QStackedBarSeries>
#include <QValueAxis>

#include <algorithm>
#include <cmath>
#include <iterator>

static const EnumStorage<QString> failStr{
    QString("Students receiving a grade of 5.0 in the closed part"),
    QString("Students receiving a grade of 5.0 in the open part"),
    QString("Students who fail the exam")};
static const EnumStorage<QString> passStr{
    QString("Students receiving a passing grade in the closed part"),
    QString("Students receiving a passing grade in the open part"),
    QString("Students who pass the exam")};

namespace impl {
qsizetype getBarIdx(const double bucketSize, const Points p,
                    const qsizetype size) noexcept {
    // The largest bucket is centered at max, so it's bucket goes up to max +
    // half a bucket size
    return std::min(size - 1, static_cast<qsizetype>(std::ceil(
                                  (p.p - bucketSize / 2.) / bucketSize)));
}
} // namespace impl

PointHistogram::PointHistogram(const GradingType gradingType,
                               const ViewSettingsWindow &viewSettings,
                               QWidget *parent)
    : QChartView(new QChart(), parent), mGradingType(gradingType),
      mSeries(new QStackedBarSeries()),
      mBarSetPassed(new QBarSet(passStr[gradingType])),
      mBarSetFailed(new QBarSet(failStr[gradingType])),
      mAxisX(new QValueAxis()), mAxisY(new QValueAxis()),
      mAxisLabelCharWidth(
          QFontMetrics(mAxisX->labelsFont()).horizontalAdvance('0')),
      mAxisLabelCharHeight(QFontMetrics(mAxisY->labelsFont()).lineSpacing()) {

    // First, set options for all chart objects
    // Bar sets
    mBarSetPassed->setColor(Qt::green);
    mBarSetPassed->setBorderColor(QColor(Qt::green).darker());
    mBarSetFailed->setColor(Qt::red);
    mBarSetFailed->setBorderColor(QColor(Qt::red).darker());

    // x-axis
    mAxisX->setLabelFormat("%d");
    mAxisX->setTickType(QValueAxis::TicksDynamic);
    mAxisX->setTickAnchor(0);

    // y-axis
    mAxisY->setLabelFormat("%d");
    mAxisY->setTickType(QValueAxis::TicksDynamic);
    mAxisY->setTickAnchor(0);

    // Stacked bar series
    // Bar width of one = no space between adjacent bars
    mSeries->setBarWidth(1.);
    mSeries->setLabelsVisible(false);

    // Legend
    chart()->legend()->setVisible(true);
    chart()->legend()->setAlignment(Qt::AlignBottom);

    // Chart
    chart()->setMinimumSize(400, 200);
    chart()->setAnimationOptions(QChart::NoAnimation);

    // Chart view
    setInteractive(false);

    //-----------------------------------------------------------------------

    // The order of connection steps is important, so that all bars are
    // distributed evenly across the whole range.
    // 1) append bar sets to series and add axis to chart
    // 2) add series to chart
    // 3) attach y axis (we intentionally skip the x axis, see rescaleX for
    // details)

    mSeries->append(mBarSetFailed);
    mSeries->append(mBarSetPassed);

    chart()->addAxis(mAxisX, Qt::AlignBottom);
    chart()->addAxis(mAxisY, Qt::AlignLeft);

    chart()->addSeries(mSeries);

    mSeries->attachAxis(mAxisY);

    //-----------------------------------------------------------------------

    // receive update when points change (rescale axis, change number of bars)
    connect(&DataManager::getInstance(), &DataManager::rawDataChanged, this,
            &PointHistogram::pointsChanged);
    // receive update when grades change (change colors)
    connect(&DataManager::getInstance(), &DataManager::gradesChanged, this,
            &PointHistogram::gradesChanged);
    // receive update when user changes bucket size
    connect(&viewSettings, &ViewSettingsWindow::pointHistoBucketSizeChanged,
            this, &PointHistogram::pointBucketSizeChanged);
}

void PointHistogram::handleResize(const QRectF plotArea) {
    mPlotArea = plotArea;

    rescaleXLabels();
    rescaleYLabels();
}

void PointHistogram::pointsChanged() {

    const auto cur = DataManager::getCurrent();
    if (!cur) {
        return;
    }

    // Overwrite the bucket size when the exam points change
    mPointBucketSize =
        std::clamp(as<double>(cur->mUserPointConfig.mGradeBoundaryStepSize),
                   defaults::minBucketSize, defaults::maxBucketSize);

    rescaleX();
    resizePointBucketLists();
    // the rest is done in the following gradesChanged-call
}
void PointHistogram::gradesChanged(const EnumStorage<bool> &types) {

    if (types[mGradingType]) {

        const auto cur = DataManager::getCurrent();
        if (!cur) {
            return;
        }

        // x axis doesn't change
        // bucket list size doesn't change
        recalculatePointBucketLists();
        updateBarSets();
        rescaleY();
    }
}

void PointHistogram::pointBucketSizeChanged(const double d) {

    mPointBucketSize = d;

    rescaleX();
    resizePointBucketLists();
    recalculatePointBucketLists();
    updateBarSets();
    rescaleY();
}

// The maximum achieved points may be larger than the "mPointsMax" because of
// bonus points
[[nodiscard]] double
getMaxPointsInclBonus(const DynexiteExam &exam,
                      const GradingType gradingType) noexcept {

    return std::max(exam.mPointsMax[gradingType].p,
                    std::ranges::max_element(exam.mPoints[gradingType])->p);
}

void PointHistogram::resizePointBucketLists() {
    const auto cur = DataManager::getCurrent();

    if (!cur) {
        mPassedStudentsPerPointBucket.resize(0);
        mFailedStudentsPerPointBucket.resize(0);

        return;
    }
    const auto &max = getMaxPointsInclBonus(cur->mExam, mGradingType);

    const auto nrBars = impl::getBarIdx(mPointBucketSize, Points{max},
                                        std::numeric_limits<qsizetype>::max()) +
                        1;
    mPassedStudentsPerPointBucket.resize(nrBars);
    mFailedStudentsPerPointBucket.resize(nrBars);
}

void PointHistogram::recalculatePointBucketLists() {
    const auto cur = DataManager::getCurrent();

    if (!cur)
        return;

    // Reset all counts to 0
    mPassedStudentsPerPointBucket.fill(0.);
    mFailedStudentsPerPointBucket.fill(0.);

    const auto &pointList = cur->mExam.mPoints[mGradingType];
    const auto &gradeList = cur->getGradeList(mGradingType);

    // For every point value, increment the correct bucket in the list of passed
    // / not-passed buckets depending on the grade
    const auto nrBars = mFailedStudentsPerPointBucket.size();
    for (int i = 0; i < pointList.size(); ++i) {
        if (gradeList[i] == Grade::grade5_0) {
            ++mFailedStudentsPerPointBucket[impl::getBarIdx(
                mPointBucketSize, pointList[i], nrBars)];
        } else {
            ++mPassedStudentsPerPointBucket[impl::getBarIdx(
                mPointBucketSize, pointList[i], nrBars)];
        }
    }
}

void PointHistogram::updateBarSets() {

    // Pausing view port update reduces flickering
    const auto currentMode = viewportUpdateMode();
    setViewportUpdateMode(NoViewportUpdate);

    {
        // Replace the bar set values with the newly calculated ones
        mBarSetPassed->remove(0, mBarSetPassed->count());
        mBarSetFailed->remove(0, mBarSetFailed->count());

        mBarSetPassed->append(mPassedStudentsPerPointBucket);
        mBarSetFailed->append(mFailedStudentsPerPointBucket);
    }

    setViewportUpdateMode(currentMode);
}

void PointHistogram::rescaleX() const {

    const auto cur = DataManager::getCurrent();

    if (!cur) {
        return;
    }

    const auto max =
        std::max(1., getMaxPointsInclBonus(DataManager::getCurrent()->mExam,
                                           mGradingType));
    // First, evenly spread the buckets over the full width
    mSeries->attachAxis(mAxisX);
    mAxisX->setRange(0, max / mPointBucketSize);

    // Then, set the axis labels to the correct values (instead of 0-"number
    // of buckets")
    mSeries->detachAxis(mAxisX);
    mAxisX->setRange(0, max);

    rescaleXLabels();
}

void PointHistogram::rescaleY() const {
    // y-axis-max is the maximum of the elementwise sum of the number of
    // students that failed/passed within each point bucket
    if (!mPassedStudentsPerPointBucket.empty()) {

        QList<qreal> totalStudentsPerPoints;
        std::ranges::transform(
            mPassedStudentsPerPointBucket, mFailedStudentsPerPointBucket,
            std::back_inserter(totalStudentsPerPoints), std::plus());

        const auto newMax = *std::ranges::max_element(totalStudentsPerPoints);

        if (mAxisY->max() != newMax) {
            mAxisY->setMax(newMax);
        }
    } else {
        mAxisY->setMax(1.);
    }

    rescaleYLabels();
}

void PointHistogram::rescaleXLabels() const {

    // Get maximum label value (roughly)
    const auto maxLabelValue = static_cast<int>(mAxisX->max());

    const auto maxLabelValueDigits =
        static_cast<int>(QString::number(maxLabelValue).length());

    // Leave one space character width around label
    const auto maxLabelPixelWidth =
        mAxisLabelCharWidth * (maxLabelValueDigits + 2);

    const auto maxNrLabels =
        std::max(2, static_cast<int>(mPlotArea.width()) / maxLabelPixelWidth);

    // We need an extra label for 0
    const auto minLabelStep = static_cast<int>(
        std::ceil(mAxisX->max() / static_cast<qreal>(maxNrLabels - 1)));

    // Round label step to next multiple of 5
    const auto labelStep =
        std::ceil(static_cast<double>(minLabelStep) / 5.) * 5.;

    mAxisX->setTickInterval(labelStep);
}

void PointHistogram::rescaleYLabels() const {

    const auto maxNrLabels = std::max(2, static_cast<int>(mPlotArea.height()) /
                                             (mAxisLabelCharHeight * 2));

    // We need an extra label for 0
    const auto minLabelStep = static_cast<int>(
        std::ceil(mAxisY->max() / static_cast<qreal>(maxNrLabels - 1)));

    mAxisY->setTickInterval(minLabelStep);
}
