#include "task_tab.hpp"

#include "data/data_manager.hpp"
#include "point_histogram.hpp"

#include <QBarCategoryAxis>
#include <QBarSeries>
#include <QBarSet>
#include <QBoxPlotSeries>
#include <QBrush>
#include <QCategoryAxis>
#include <QChart>
#include <QChartView>
#include <QLabel>
#include <QValueAxis>

#include <algorithm>
#include <iomanip>
#include <sstream>

TaskTab::TaskTab(QWidget *parent) : QScrollArea(parent) {

    setWindowTitle("Tasks");
    setWidgetResizable(true);

    const auto contents = new QWidget(this);
    mLayout = new QGridLayout(contents);
    setWidget(contents);

    connect(&DataManager::getInstance(), &DataManager::rawDataChanged, this,
            &TaskTab::pointsChanged);
}

void TaskTab::pointsChanged() {

    // Delete previous charts
    {
        delete mLayout;
        // The following deletes all child widgets, i.e. the charts and
        // chartViews
        delete takeWidget();

        const auto contents = new QWidget(this);
        mLayout = new QGridLayout(contents);
        setWidget(contents);
    }

    const auto cur = DataManager::getCurrent();
    if (!cur) {
        return;
    }

    const auto &taskEarnedPoints =
        cur->mExam.mTaskEarnedPoints[GradingType::Combined];
    const auto &taskMaxPoints = cur->mExam.mTaskMaxPoints;
    const auto &taskNames = cur->mExam.mTaskNames;

    for (int i = 0; i < taskEarnedPoints.size(); ++i) {

        // Get all points earned in the task
        auto sortedPoints = taskEarnedPoints[i];
        std::ranges::sort(sortedPoints);

        const auto nrStudents = taskEarnedPoints[i].size();

        // Special case bonus points: hide if no students has bonus points
        if (i == taskEarnedPoints.size() - 1 &&
            taskNames[i] == "Bonus points" && nrStudents == 0) {
            continue;
        }

        // Task  name
        {
            const auto nameLabel = new QLabel(taskNames[i]);
            nameLabel->setWordWrap(true);
            mLayout->addWidget(nameLabel, i * 3, 0, 2, 1);
        }

        if (nrStudents == 0) {

            mLayout->addWidget(
                new QLabel("No student participated in this task"), i * 3, 1, 2,
                1);
        } else {

            // Value of the highest label on the y axis of the point chart.
            // Used to determine how much space to leave to the left of the
            // box-and-whiskers-plot
            int pointChartYMax = 0;

            const auto stepSize =
                as<qreal>(cur->mUserPointConfig.mGradeBoundaryStepSize);
            const auto maxPoints =
                std::max(sortedPoints.back().p, taskMaxPoints[i].p);
            const auto nrBuckets =
                static_cast<int>(std::ceil(maxPoints / stepSize) + 1);

            // Point bucket plot
            {
                auto pointChart = new QChart;

                // For each possible point value, count how many students
                // got that many points
                QList<qreal> nrStudentsPerPoint;
                nrStudentsPerPoint.resize(nrBuckets);

                for (const auto p : taskEarnedPoints[i]) {
                    ++nrStudentsPerPoint[impl::getBarIdx(stepSize, p,
                                                         nrBuckets)];
                }

                auto *taskSeries = new QBarSeries();
                taskSeries->setBarWidth(1.);
                taskSeries->setLabelsVisible(false);

                auto *set = new QBarSet("");
                set->append(nrStudentsPerPoint);

                taskSeries->append(set);

                pointChart->addSeries(taskSeries);

                // x-axis
                {
                    const auto axisX = new QBarCategoryAxis();

                    for (int j = 0; j < nrBuckets; ++j) {
                        std::stringstream category;
                        category << std::fixed
                                 << std::setprecision(nrDecimalPlaces(
                                        cur->mUserPointConfig
                                            .mGradeBoundaryStepSize))
                                 << (j * static_cast<double>(stepSize));
                        axisX->append(QString::fromStdString(category.str()));
                    }
                    pointChart->addAxis(axisX, Qt::AlignBottom);
                    taskSeries->attachAxis(axisX);
                }

                // y-axis
                {
                    const auto largestBucket =
                        *std::ranges::max_element(nrStudentsPerPoint);

                    const auto axisY = new QValueAxis();
                    axisY->setTickType(QValueAxis::TicksDynamic);
                    axisY->setLabelFormat("%d");
                    axisY->setTickAnchor(0);

                    axisY->setMin(0.0);

                    // Find sensible tick marks
                    static constexpr int maxTicks = 8;
                    if (static_cast<int>(largestBucket) <= maxTicks) {

                        axisY->setTickInterval(1.);
                        pointChartYMax = static_cast<int>(largestBucket);
                        axisY->setMax(pointChartYMax);
                    } else {
                        // Find ideal number of ticks
                        int bestTickNr = 0;
                        int bestTickNrMax = std::numeric_limits<int>::max();

                        for (int tickCount = maxTicks; tickCount > maxTicks - 3;
                             --tickCount) {
                            // Find smallest multiple of tickCount larger
                            // than largestBucket
                            const auto potentialMax =
                                static_cast<int>(
                                    std::ceil(largestBucket / tickCount)) *
                                tickCount;

                            if (potentialMax - static_cast<int>(largestBucket) <
                                bestTickNrMax -
                                    static_cast<int>(largestBucket)) {

                                bestTickNr = tickCount;
                                bestTickNrMax = potentialMax;
                            }
                        }

                        // bestTickNrMax is a multiple of bestTickNr, so the
                        // result of this is an integer
                        axisY->setTickInterval(
                            static_cast<qreal>(bestTickNrMax / bestTickNr));

                        pointChartYMax = bestTickNrMax;
                        axisY->setMax(pointChartYMax);
                    }

                    pointChart->addAxis(axisY, Qt::AlignLeft);
                    taskSeries->attachAxis(axisY);
                }

                pointChart->setAnimationOptions(QChart::NoAnimation);
                pointChart->legend()->hide();

                const auto chartView = new QChartView(pointChart);
                chartView->setMinimumSize(400, 300);
                chartView->setInteractive(false);

                mLayout->addWidget(chartView, i * 3, 1);
            }

            // BoxPlot
            {
                auto boxChart = new QChart;

                auto *taskSeries = new QBoxPlotSeries;
                auto *set = new QBoxSet;

                // Make sure the different plot elements don't overlap each
                // other
                // const qreal minDistance = static_cast<qreal>(nrBuckets) *
                // 0.005;
                const qreal minDistance = stepSize * 0.1;

                const qreal median = sortedPoints[nrStudents / 2].p;
                const qreal lowerQuartile = std::min(
                    sortedPoints[nrStudents / 4].p, median - minDistance);
                const qreal lowerExtreme = std::min(
                    sortedPoints.begin()->p, lowerQuartile - minDistance);
                const qreal upperQuartile = std::max(
                    sortedPoints[3 * nrStudents / 4].p, median + minDistance);
                const qreal upperExtreme = std::max(
                    sortedPoints.back().p, upperQuartile + minDistance);

                set->setValue(QBoxSet::Median, median);
                set->setValue(QBoxSet::LowerQuartile, lowerQuartile);
                set->setValue(QBoxSet::UpperQuartile, upperQuartile);
                set->setValue(QBoxSet::LowerExtreme, lowerExtreme);
                set->setValue(QBoxSet::UpperExtreme, upperExtreme);

                taskSeries->setBrush(QBrush(Qt::GlobalColor::green));
                taskSeries->append(set);

                boxChart->addSeries(taskSeries);

                // Add a y axis to get the same offset at the left as the point
                // histogram, for easy point alignment
                {
                    const auto axisY = new QCategoryAxis();
                    axisY->append(QString::number(pointChartYMax), 0);
                    // Set label color to white so that it is hopefully
                    // invisible
                    axisY->setLabelsColor(QColorConstants::White);

                    // chart is rotated 90 degrees, so this is on the left
                    axisY->setLabelsAngle(270);
                    boxChart->addAxis(axisY, Qt::AlignBottom);

                    taskSeries->attachAxis(axisY);
                }

                // add x axis to specify min and max
                {
                    const auto axisX = new QValueAxis();
                    // Set label format so that the label is invisible
                    axisX->setLabelFormat(" ");
                    axisX->setRange(0. - stepSize / 2.,
                                    maxPoints + stepSize / 2.);
                    axisX->setTickType(QValueAxis::TicksFixed);
                    axisX->setTickCount(2);

                    // chart is rotated 90 degrees, so this is on the left
                    axisX->setLabelsAngle(270);
                    boxChart->addAxis(axisX, Qt::AlignRight);
                    taskSeries->attachAxis(axisX);
                }

                boxChart->setAnimationOptions(QChart::NoAnimation);
                boxChart->legend()->hide();

                const auto chartView = new QChartView(boxChart);
                chartView->setMinimumSize(400, 150);
                chartView->setInteractive(false);

                mLayout->addWidget(chartView, i * 3 + 1, 1);

                chartView->rotate(90);

                // The QT box plot has a little bit more space above it (because
                // of our rotation, to the right of it) than the bar chart has
                // to the right of it. Therefore, reduce the upper (because of
                // our rotation, right) margin by 6px.
                // Furthermore, reduce the left and right because of
                // our rotation, above and below) margin to have less space
                // between box plot and bar chart.
                boxChart->setMargins(QMargins(5, 14, 0, 20));
            }
        }

        // Separator line
        {
            const auto line = new QFrame;
            line->setFrameShape(QFrame::HLine);
            line->setFrameShadow(QFrame::Sunken);
            line->setMinimumHeight(30);

            mLayout->addWidget(line, i * 3 + 2, 0, 1, 2);
        }
    }

    // Ratio of 1:10 between name and chart width
    mLayout->setColumnStretch(0, 1);
    mLayout->setColumnStretch(1, 10);

    // Reduce space between box plot and bar chart
    mLayout->setVerticalSpacing(0);
}
